;=====================================================================================
; Texture Map a triangle (Affine Texture Mapping)
;=====================================================================================
; The alias: anytriangle
; This takes <x y> <x y> <x y> <u v> <u v> <u v>
; The three X,Y coordinates define the three points that make up the triangle
; The three U,V coordinates define the relation to the texture
;
; This assumes that the triangle is one of the three types of triangles, flat top
; flat bottom, or neither (which is really both)
;=====================================================================================
; Some basic examples of using two triangles to form a square
;=====================================================================================
alias triexample1 {
  if ($window(@Triangle)) { clear @triangle }
  anytriangle 210 130 430 350 210 350 0 0 179 179 0 179
  anytriangle 210 130 430 350 430 130 0 0 179 179 179 0
}
alias triexample2 {
  if ($window(@Triangle)) { clear @triangle }
  anytriangle 158 240 393 386 264 353 0 0 179 179 0 179
  anytriangle 158 240 393 386 282 240 0 0 179 179 179 0
}
alias triexample3 {
  if ($window(@Triangle)) { clear @triangle }
  anytriangle 247 94 482 240 358 240 0 0 179 179 0 179
  anytriangle 247 94 482 240 376 127 0 0 179 179 179 0
}
alias triexample4 {
  if ($window(@Triangle)) { clear @triangle }
  anytriangle 320 85 320 395 165 240 0 0 179 179 0 179
  anytriangle 320 85 320 395 475 240 0 0 179 179 179 0
}

;=====================================================================================
; More advanced example, texture mapping a 3D cube
;=====================================================================================

menu @Triangle {
  mouse {
    if ($mouse.key & 1) {
      var %cx = $window(@Triangle).dw / 2 , %cy = $window(@Triangle).dh / 2
      var %c.x = $calc(360 * ($mouse.y - %cy) / %cy)) , %c.y = $calc(-360 * (($mouse.x - %cx) / %cx)) , %c.z = $calc(360 * ($mouse.y - %cy) / %cy))
      TriCube -l %c.x %c.y
      .timerDrawTriTex -m 1 500 TriCube %c.x %c.y
    }
  }
}
alias TriCube {
  if ($1 = -l) { var %linesonly = $true | tokenize 32 $2- }
  if (!$window(@Triangle)) { window -dfp @Triangle -1 -1 640 480 }
  if (!$window(@TriTex)) { window -dfp @TriTex -1 -1 180 180 | drawpic @TriTex 0 0 $qt($scriptdir\texmap1.bmp) }
  drawrect -rnf @Triangle 2105376 0 0 0 $window(@Triangle).dw $window(@Triangle).dh

  var %cx = $window(@Triangle).dw / 2 , %cy = $window(@Triangle).dh / 2 , %Ticks = $ticks
  var %3D.1 = -90 -90 -90 , %3D.2 = 90 -90 -90 , %3D.3 = 90 90 -90 , %3D.4 = -90 90 -90
  var %3D.5 = -90 -90 90 , %3D.6 = 90 -90 90 , %3D.7 = 90 90 90 , %3D.8 = -90 90 90
  ;var %Tables = 1 2 3 4 1;2 6 7 3 2;6 5 8 7 6;5 1 4 8 5;5 6 2 1 5;4 3 7 8 4
  var %Tables = 1 2 3 1;1 3 4 1;2 6 7 2;2 7 3 2;6 5 8 6;6 8 7 6;5 1 4 5;5 4 8 5;5 6 2 5;5 2 1 5;4 3 7 4;4 7 8 4
  var %UVTables = 0 0 179 0 179 179;0 0 179 179 0 179;0 0 179 0 179 179;0 0 179 179 0 179;0 0 179 0 179 179;0 0 179 179 0 179;0 0 179 0 179 179;0 0 179 179 0 179;0 0 179 0 179 179;0 0 179 179 0 179;0 0 179 0 179 179;0 0 179 179 0 179;
  var %xr = $iif($1,$1,0) , %yr = $iif($2,$2,0) , %zr = $iif($3,$3,0)
  var %x = 8
  while (%x) { 
    tokenize 32 %xr %yr %zr $var($+(%,3D.,%x),1).value
    tokenize 32 $1-3 $4 $calc(($5 * $cos($1).deg) + ($6 * (0 - $sin($1).deg))) $calc(($5 * $sin($1).deg) + ($6 * $cos($1).deg))
    tokenize 32 $1-3 $calc(($4 * $cos($2).deg) + ($6 * $sin($2).deg)) $5 $calc(($4 * (0 - $sin($2).deg)) + ($6 * $cos($2).deg))
    tokenize 32 $calc(($4 * $cos($3).deg) + ($5 * (0 - $sin($3).deg))) $calc(($4 * $sin($3).deg) + ($5 * $cos($3).deg)) $6
    var %C3D. $+ %x $1-
    dec %x
  }
  var %asdx = 12
  while (%asdx) {
    var %table = $gettok(%Tables,%asdx,59) , %y = $numtok(%table,32) , %x = %y - 1
    tokenize 32 $Tri.convert2d(%cx,%cy,$var($+(%,C3D.,$gettok(%table,1,32)),1).value) $Tri.convert2d(%cx,%cy,$var($+(%,C3D.,$gettok(%table,2,32)),1).value) $Tri.convert2d(%cx,%cy,$var($+(%,C3D.,$gettok(%table,3,32)),1).value) $Tri.convert2d(%cx,%cy,$var($+(%,C3D.,$gettok(%table,4,32)),1).value)
    var %cull = $calc(($1 - $3) * ($6 - $4) - ($2 - $4) * ($5 - $3)) , %avgx = $calc(($1 + $3 + $5) / 3) , %avgy = $calc(($2 + $4 + $6) / 3)
    if (%cull < 0) { 
      if (%linesonly) { 
        drawline -n @Triangle 1 1 $1- 
        drawtext -n @Triangle 15 %avgx %avgy %asdx
      }
      else { 
        if (%asdx isnum 1-2) { drawpic @TriTex 0 0 $qt($scriptdir\texmap1.bmp) }
        if (%asdx isnum 3-4) { drawpic @TriTex 0 0 $qt($scriptdir\texmap2.bmp) }
        if (%asdx isnum 5-6) { drawpic @TriTex 0 0 $qt($scriptdir\texmap3.bmp) }
        if (%asdx isnum 7-8) { drawpic @TriTex 0 0 $qt($scriptdir\texmap4.bmp) }
        if (%asdx isnum 9-10) { drawpic @TriTex 0 0 $qt($scriptdir\texmap5.bmp) }
        if (%asdx isnum 11-12) { drawpic @TriTex 0 0 $qt($scriptdir\texmap6.bmp) }
        anytriangle $1-6 $gettok(%UVTables,%asdx,59) 
      }
    }
    dec %asdx
  }
  drawdot @Triangle
  var %ms = $ticks - %ticks
  titlebar @Triangle Rendered in: %ms ms $round($calc(1000 / %ms),2) FPS
}
alias Tri.convert2d {
  var %cx = $1 , %cy = $2
  tokenize 32 $3-
  return $round($calc(%cx + ($1 * 500) / ($3 + 500)),0) $round($calc(%cy + ($2 * 500) / ($3 + 500)),0)
}

;=====================================================================================
; The alias itself in all its glory! anytriangle!
;=====================================================================================

alias anytriangle {
  if (!$window(@Triangle)) { window -dfp @Triangle -1 -1 640 480 }
  if (!$window(@TriTex)) { window -dfp @TriTex -1 -1 180 180 | drawpic @TriTex 0 0 $qt($scriptdir\texmap1.bmp) }

  ;Store all the coordinates for easy sorting
  var %irestart = left
  var %x0 = $1 , %y0 = $2 , %x1 = $3 , %y1 = $4 , %x2 = $5 , %y2 = $6
  var %u0 = $7 , %v0 = $8 , %u1 = $9 , %v1 = $10 , %u2 = $11 , %v2 = $12

  ;Sort all the coordinates to get a positive direction
  if (%y1 < %y0) {
    var %tempx = %x1 , %tempy = %y1 , %tempu = %u1 , %tempv = %v1
    %x1 = %x0
    %y1 = %y0
    %u1 = %u0
    %v1 = %v0

    %x0 = %tempx
    %y0 = %tempy
    %u0 = %tempu
    %v0 = %tempv
  }
  if (%y2 < %y0) {
    var %tempx = %x2 , %tempy = %y2 , %tempu = %u2 , %tempv = %v2
    %x2 = %x0
    %y2 = %y0
    %u2 = %u0
    %v2 = %v0

    %x0 = %tempx
    %y0 = %tempy
    %u0 = %tempu
    %v0 = %tempv
  }
  if (%y2 < %y1) {
    var %tempx = %x2 , %tempy = %y2 , %tempu = %u2 , %tempv = %v2
    %x2 = %x1
    %y2 = %y1
    %u2 = %u1
    %v2 = %v1

    %x1 = %tempx
    %y1 = %tempy
    %u1 = %tempu
    %v1 = %tempv
  }
  ;test for a flat top triangle
  if (%y0 = %y1) { 
    var %tritype = flattop 

    ;sort coordinates left to right
    if (%x1 < %x0) {
      var %tempx = %x1 , %tempy = %y1 , %tempu = %u1 , %tempv = %v1
      %x1 = %x0
      %y1 = %y0
      %u1 = %u0
      %v1 = %v0

      %x0 = %tempx
      %y0 = %tempy
      %u0 = %tempu
      %v0 = %tempv
    }
  }
  ;test for a flat bottom triangle
  elseif (%y1 = %y2) {
    var %tritype = flatbottom

    ;sort coordinates left to right
    if (%x2 < %x1) {
      var %tempx = %x2 , %tempy = %y2 , %tempu = %u2 , %tempv = %v2
      %x2 = %x1
      %y2 = %y1
      %u2 = %u1
      %v2 = %v1

      %x1 = %tempx
      %y1 = %tempy
      %u1 = %tempu
      %v1 = %tempv
    }
  }
  else {
    var %tritype = neither  
  }

  if (%tritype != neither) {
    if (%tritype = flattop) {
      var %dy = $calc(%y2 - %y0)
      var %dxdyl = $calc((%x2 - %x0) / %dy)
      var %dudyl = $calc((%u2 - %u0) / %dy)
      var %dvdyl = $calc((%v2 - %v0) / %dy)

      var %dxdyr = $calc((%x2 - %x1) / %dy)
      var %dudyr = $calc((%u2 - %u1) / %dy)
      var %dvdyr = $calc((%v2 - %v1) / %dy)

      var %xl = %x0
      var %xr = %x1

      var %ul = %u0
      var %ur = %u1

      var %vl = %v0
      var %vr = %v1

      var %ystart = %y0
    }
    elseif (%tritype = flatbottom) {
      var %dy = $calc(%y1 - %y0)
      var %dxdyl = $calc((%x1 - %x0) / %dy)
      var %dudyl = $calc((%u1 - %u0) / %dy)
      var %dvdyl = $calc((%v1 - %v0) / %dy)

      var %dxdyr = $calc((%x2 - %x0) / %dy)
      var %dudyr = $calc((%u2 - %u0) / %dy)
      var %dvdyr = $calc((%v2 - %v0) / %dy)

      var %xl = %x0
      var %xr = %x0

      var %ul = %u0
      var %ur = %u0

      var %vl = %v0
      var %vr = %v0

      var %ystart = %y0  
    }
    var %y = %ystart , %yend = %y2
    while (%y <= %yend) {
      var %xstart = %xl , %xend = %xr
      var %ui = %ul
      var %vi = %vl
      var %dx = $calc(%xend - %xstart)
      if (%dx > 0) {
        var %du = $calc((%ur - %ul) / %dx)
        var %dv = $calc((%vr - %vl) / %dx)
      }
      else {
        var %du = $calc(%ur - %ul)
        var %dv = $calc(%vr - %vl)
      }
      var %x = %xstart
      while (%x <= %xend) {
        var %color = $getdot(@tritex,%ui,%vi)
        drawdot -r @triangle %color 1 %x %y
        inc %ui %du
        inc %vi %dv
        inc %x
      }
      inc %xl %dxdyl
      inc %ul %dudyl
      inc %vl %dvdyl

      inc %xr %dxdyr
      inc %ur %dudyr
      inc %vr %dvdyr

      inc %y
    }
  }
  else {
    var %yend = %y2 , %yrestart = %y1

    var %dyl = $calc(%y1 - %y0)
    var %dxdyl = $calc((%x1 - %x0) / %dyl)
    var %dudyl = $calc((%u1 - %u0) / %dyl)
    var %dvdyl = $calc((%v1 - %v0) / %dyl)

    var %dyr = $calc(%y2 - %y0)
    var %dxdyr = $calc((%x2 - %x0) / %dyr)
    var %dudyr = $calc((%u2 - %u0) / %dyr)
    var %dvdyr = $calc((%v2 - %v0) / %dyr)

    var %xl = %x0
    var %xr = %x0

    var %ul = %u0
    var %ur = %u0

    var %vl = %v0
    var %vr = %v0

    var %ystart = %y0  
    ;test swap to keep render left to right
    if (%dxdyr < %dxdyl) {
      var %temp = %dxdyl , %dxdyl = %dxdyr , %dxdyr = %temp
      var %temp = %dudyl , %dudyl = %dudyr , %dudyr = %temp
      var %temp = %dvdyl , %dvdyl = %dvdyr , %dvdyr = %temp

      var %temp = %xl , %xl = %xr , %xr = %temp
      var %temp = %ul , %ul = %ur , %ur = %temp
      var %temp = %vl , %vl = %vr , %vr = %temp

      var %temp = %x1 , %x1 = %x2 , %x2 = %temp
      var %temp = %y1 , %y1 = %y2 , %y2 = %temp

      var %temp = %u1 , %u1 = %u2 , %u2 = %temp
      var %temp = %v1 , %v1 = %v2 , %v2 = %temp
      %irestart = right
    }

    var %y = %ystart
    while (%y <= %yend) {
      var %xstart = %xl
      var %xend = %xr
      var %ui = %ul
      var %vi = %vl

      var %dx = $calc(%xend - %xstart)
      if (%dx > 0) {
        var %du = $calc((%ur - %ul) / %dx)
        var %dv = $calc((%vr - %vl) / %dx)
      }
      else {
        var %du = $calc(%ur - %ul)
        var %dv = $calc(%vr - %vl)
      }
      var %x = %xstart
      while (%x <= %xend) {
        var %color = $getdot(@tritex,%ui,%vi)
        drawdot -r @triangle %color 1 %x %y
        inc %ui %du
        inc %vi %dv    
        inc %x
      }
      if (%y = %yrestart) {
        if (%irestart = left) {
          var %dyl = $calc(%y2 - %y1)
          var %dxdyl = $calc((%x2 - %x1) / %dyl)
          var %dudyl = $calc((%u2 - %u1) / %dyl)
          var %dvdyl = $calc((%v2 - %v1) / %dyl)

          var %xl = %x1
          var %ul = %u1
          var %vl = %v1     
        }
        else {
          var %dyr = $calc(%y1 - %y2)
          var %dxdyr = $calc((%x1 - %x2) / %dyr)
          var %dudyr = $calc((%u1 - %u2) / %dyr)
          var %dvdyr = $calc((%v1 - %v2) / %dyr)

          var %xr = %x2
          var %ur = %u2
          var %vr = %v2
        }
      }
      inc %xl %dxdyl
      inc %ul %dudyl
      inc %vl %dvdyl

      inc %xr %dxdyr
      inc %ur %dudyr
      inc %vr %dvdyr
      inc %y
    }
  }
}
